#!/usr/bin/env python

## Generate a heatmap based on the event database
#
# Usage:
# 
# $ ./clean_db
# $ ./import_source /var/log/log.type.1
# $ ./import_source /var/log/log.type.2
# $ ./generate_heatmap my_heatmap.png
#
# The output file is produced in a cylindrical equirectangular projection,
# and is suitable for use in tools such as Google Earth. A sample .kml file
# is provided under the examples/ directory.
#
# This is released under the terms of the BSD license. See COPYRIGHT for
# further information.
#

import sqlite3
import sys
import numpy
import GeoIP
import Image
import getopt
from eventdb import EventDB


# Default to the Pitcairn Islands when we don't have any real location for the
# IP. It's a shame we can't produce a height map too, and put it on the ISS's
# orbit.
DEFAULT_LOCATION = (-24.3742, -128.3156)


## Generate the heat map colour lookup table
#
# This routine generates an 8-bit LUT for the heat map, with the first
# entry (no events) being completely transparent, and the remainder are
# opaque with a steady gradient from yellow to red.
#
# @return  Returns a 256*4 array, with each LUT entry in RGBA order.
#
def generate_colour_lut():
    lut = numpy.zeros((256,4), numpy.uint8)

    for x in xrange(1,256):
        lut[x][0] = 255        # Red channel
        lut[x][1] = 256 - x    # Green channel
        # Blue channel is already zero
        lut[x][3] = 255        # Alpha channel

    return lut


## Generate a heatmap based on the event count database
#
# This routine generates a geographic heatmap in a really basic projection -
# we use a simple equirectangular projection, where each pixel represents a
# fixed angular difference. 
#
# The Prime Meridian (0-degree meridian) is located at the horizontal centre
# of the image, and the Equator (0-degree parallel) is located at the vertical
# centre of the image.
#
# While a simple projection, this has the benefit that most other GIS
# developers have been equally lazy, which permits us to use this heatmap
# unaltered as a texture (overlay image) in Google Earth for end-user
# visualisation purposes.
#
# We produce a heatmap that contains all 360 degrees of longitude, and all
# 180 degrees of latitude.  The heatmap will be saved to the file indicated
# in 'output_filename'.
#
# @param  output_filename  File to write the heatmap image to
# @param  x_res            Resolution in the longitudinal axis, in pixels.
# @param  y_res            Resolution in the latitudal axis, in pixels.
# @param  geoipDB          Filename for the GeoIP database, which must be
#                          provided in binary format.
# @param  eventsDB         Filename for the parsed events SQLite database.
#
def generate_heatmap(output_filename, x_res = 720, y_res = 360, geoipDB = 'GeoLiteCity.dat', eventsDB = 'events.sqlite'):

    # Grab the event database
    db = sqlite3.connect(eventsDB)

    # Grab the GeoIP database for lat/long correlation
    gi = GeoIP.open(geoipDB, GeoIP.GEOIP_STANDARD)

    # Generate the raw buckets to store the event counts in
    rawData = numpy.zeros((y_res, x_res))

    # Go through the events and throw the counts into the buckets
    c = db.cursor()
    c.execute('SELECT ipNum, count FROM events ORDER BY ipNum')
   
    for row in c:
        ip, count = row

        # Look up the lat/long for the IP
        info = gi.record_by_addr(EventDB.int_to_ip(ip))

        # Work out which bucket to throw it in
        if info is not None:
            latitude  = round(( 90 - info['latitude'])  * y_res / 180)
            longitude = round((180 + info['longitude']) * x_res / 360)
        else:
            (latitude, longitude) = DEFAULT_LOCATION

        # Throw the data into the bucket
        rawData[latitude, longitude] += count

    # Because we have some locations with a lot of outliers, we want to 
    # correct the data to be non-linear for user display. Scale the data
    # by a power of 0.4, and find the maximum.
    #
    # The power of 0.4 is quite arbitrary, but I found it fit smaller
    # data sets quite well. This avoids needing to use a histogram or
    # other methods to classify the data, at the expense of accurately
    # conveying unusual distributions.
    #
    rawData = pow(rawData, 0.4)

    # Find the maximum scaled value, and scale all the data between 0 and 255.
    maxVal = max([1, numpy.max(rawData)])
    rawData = rawData * (255.0 / maxVal)

    # Now transform the data for output display. We want to show locations
    # with no events as transparent, and use a colour look-up table to 
    # display the number of events otherwise.
    #
    lut = generate_colour_lut()
    rawData = numpy.round(rawData)
    intData = numpy.array(rawData, numpy.uint8)
    colData = lut[intData]

    # Convert the heatmap to an image, and dump it out. PNG is suggested.
    #
    im = Image.fromarray(colData, mode='RGBA')
    im.save(output_filename)


## Program usage information
#
def usage(progName):
    print >>sys.stderr, "generate_heatmap v0.1"
    print >>sys.stderr, "Usage:\n"
    print >>sys.stderr, "  %s [option] ... output.png" % progName
    print >>sys.stderr, ""
    print >>sys.stderr, "  -x x_res                 Create a heatmap that is x_res pixels wide."
    print >>sys.stderr, "  -y y_res                 Create a heatmap that is y_res pixels high."
    print >>sys.stderr, "  --geoip-db geo.dat       Use geo.dat for the IP lookup table."
    print >>sys.stderr, "  --events-db events.sl3   Use events.sl3 as the source for the heatmap data."
    print >>sys.stderr, ""


## Main entry-point when called as a script
#
def main(argv):
    try:    
        # Parse arguments with getopt
        options, trailer = getopt.getopt(argv[1:], 'x:y:', ['geoip-db=', 'events-db='])

        # Check for erroneous filenames
        if len(trailer) != 1:
            raise Exception('You must specify a single filename to save the heatmap to.')

        numArgs = {'-x': ('x_res', 'You must specify a positive integer for the X-resolution.'),
                   '-y': ('y_res', 'You must specify a positive integer for the Y-resolution.')}
        fileArgs = {'--geoip-db':  'geoipDB',
                    '--events-db': 'eventsDB'}
        args = {}

        for option, value in options:
            if option in numArgs:

                # Extract the parameter and error message
                param, message = numArgs[option]
                try:
                    i = int(value)

                    # Ensure they provide a sane value.
                    if i < 1:
                        raise ValueError()
                    else:
                        args[param] = i

                except ValueError, e:
                    raise ValueError(message)

            elif option in fileArgs:
                
                # Extract the parameter name
                param = fileArgs[option]
                if not os.path.isfile(value):
                    raise Exception('%s is not a file.' % value)

                args[param] = value
            else:
                raise Exception('Unknown option "%s".' % option)

        generate_heatmap(trailer[0], **args)

    except Exception, e:
        print e
        print >>sys.stderr, "Error: %s" % e.message
        print >>sys.stderr, ""
        usage(argv[0])
        exit(1)


# If run as a standalone app, behave accordingly.
#
if __name__ == '__main__':
    main(sys.argv)

